"""
========================================================================
|                                                                      |
|                       SUSYAI EXAMPLE SCRIPT 02:                      |
|                   PREDICTION OF MULTIPLE COORDINATE                  |
|   ----------------------------------------------------------------   |
|                                                                      |
|   SUSY-AI is a powerful, yet simple interface between SUSY           |
|   data and the sklearn Python class. With the classifier that        |
|   comes with SUSY-AI, it allows for predicting of exclusion of a     |
|   datapoint in pMSSM parameterspace via sklearn using a              |
|   pre-trained classifier in only a couple of commands.               |
|                                                                      |
|   Full documentation, and the most recent version of the software    |
|   can be found on the project website                                |
|                                                                      |
|       https://susyai.hepforge.org/                                   |
|                                                                      |
|   If you are using this script scientifically or commercially,       |
|   please reference the following paper                               |
|                                                                      |
|       Sascha Caron, Jong Soo Kim, Krzysztof Rolbiecki, Roberto       |
|       Ruiz de Austri, Bob Stienen, 'The BSM-AI project: SUSY-AI -    |
|       Generalizing LHC limits on Supersymmetry with Machine          |
|       Learning'. arXiv: 1605.02797 [hep-ph]                          |
|                                                                      |
|   For questions about the script, contact                            |
|                                                                      |
|       Bob Stienen <b.stienen@student.ru.nl>                          |
|                                                                      |
========================================================================
"""


from susyai import susyai
import numpy as np

# Defining the coordinate
coordinate = np.array([[ 811.65,	# M1
	-526.38,		# M2
	2339.95,		# M3
	3421.54,		# mL1
	2861.32,		# mL3
	3496.63,		# me1
	3339.50,		# me3
	1994.02,		# mQ1
	2230.6599,		# mQt
	2691.41,		# mu1
	1842.7199,		# mtR
	3941.68,		# md1
	1799.2599,		# mbR
	-3041.83,		# At
	-3501.52,		# Ab
	-1533.57,		# Atau
	351.21,			# mu
	3423.61,		# MA
	12.25],			# tanbeta
	[-3800.69, 2881.01, 1688.58, 1020.47, 2283.25, 1888.7, 1047.99, 3108.86, 2056.61, 3898.88, 1712.05, 2943.75, 1763.34, 2629.84, -2241.57, -3969.9, -813.08, 2199.54, 10.56]])


# Creating susyai object
sa = susyai()

# Set the location of the pickle to a valid location before testing
# Loading before prediction removes the need for it during prediction, thus speeding it up
sa.set_pickle_location("pmssm.pkl")
sa.load_pickle()


# Prediction
clas, pred, conf, coords = sa.predict(coordinate)

print("classification: ",clas)
print("prediction:     ",pred)
print("confidence:     ",conf)
