"""
========================================================================
|                                                                      |
|                       SUSYAI EXAMPLE SCRIPT 12:                      |
|                           MINIMAL CERTAINTY                          |
|   ----------------------------------------------------------------   |
|                                                                      |
|   SUSY-AI is a powerful, yet simple interface between SUSY           |
|   data and the sklearn Python class. With the classifier that        |
|   comes with SUSY-AI, it allows for predicting of exclusion of a     |
|   datapoint in pMSSM parameterspace via sklearn using a              |
|   pre-trained classifier in only a couple of commands.               |
|                                                                      |
|   Full documentation, and the most recent version of the software    |
|   can be found on the project website                                |
|                                                                      |
|       https://susyai.hepforge.org/                                   |
|                                                                      |
|   If you are using this script scientifically or commercially,       |
|   please reference the following paper                               |
|                                                                      |
|       Sascha Caron, Jong Soo Kim, Krzysztof Rolbiecki, Roberto       |
|       Ruiz de Austri, Bob Stienen, 'The BSM-AI project: SUSY-AI -    |
|       Generalizing LHC limits on Supersymmetry with Machine          |
|       Learning'. arXiv: 1605.02797 [hep-ph]                          |
|                                                                      |
|   For questions about the script, contact                            |
|                                                                      |
|       Bob Stienen <b.stienen@student.ru.nl>                          |
|                                                                      |
========================================================================
"""

from susyai import susyai
import numpy as np
from os import popen

# Define files
files = []
fs = popen('ls %s | grep "%s"' % ('spectrumfiles', '.slha')).readlines()
for j in range(len(fs)):
	files.append('spectrumfiles/'+fs[j].rstrip())


# Creating susyai object
sa = susyai()

# Set the location of the pickle to a valid location before testing
# Loading before prediction removes the need for it during prediction, thus speeding it up
sa.set_pickle_location("pmssm.pkl")
sa.load_pickle()

# Set coordinate selection to SLHA-1 standard
sa.set_coordinate_selector(1)

# Set mapping of coordinates (default: False)
sa.set_map_outsiders(True)

# Prediction with a minimal confidence of 0.95
# Note that variable COORDS now containes the mapped coordinates!
clas, pred, conf, coords = sa.predict_files(files, min_confidence = 0.95)

# Compare the results to those of ex11!
print("classification: ",clas)
print("prediction:     ",pred)
print("confidence:     ",conf)
