"""
========================================================================
|                                                                      |
|                       SUSYAI EXAMPLE SCRIPT 15:                      |
|                          GET DATA FROM FILE                          |
|   ----------------------------------------------------------------   |
|                                                                      |
|   SUSY-AI is a powerful, yet simple interface between SUSY           |
|   data and the sklearn Python class. With the classifier that        |
|   comes with SUSY-AI, it allows for predicting of exclusion of a     |
|   datapoint in pMSSM parameterspace via sklearn using a              |
|   pre-trained classifier in only a couple of commands.               |
|                                                                      |
|   Full documentation, and the most recent version of the software    |
|   can be found on the project website                                |
|                                                                      |
|       https://susyai.hepforge.org/                                   |
|                                                                      |
|   If you are using this script scientifically or commercially,       |
|   please reference the following paper                               |
|                                                                      |
|       Sascha Caron, Jong Soo Kim, Krzysztof Rolbiecki, Roberto       |
|       Ruiz de Austri, Bob Stienen, 'The BSM-AI project: SUSY-AI -    |
|       Generalizing LHC limits on Supersymmetry with Machine          |
|       Learning'. arXiv: 1605.02797 [hep-ph]                          |
|                                                                      |
|   For questions about the script, contact                            |
|                                                                      |
|       Bob Stienen <b.stienen@student.ru.nl>                          |
|                                                                      |
========================================================================
"""

from susyai import susyai
import numpy as np

# Creating susyai object
sa = susyai()

# Set coordinate selector to preset-1
sa.set_coordinate_selector(1)

# Set file id selector
sa.set_id_selector("fullfilename")

# Get data from file
coords, xtras, fileid = sa.get_data_from_file('spectrumfiles/spectrum_01.slha')

# No xtra_selector and id_selector have been set, so they are an empty array and list respectively
print("coordinates: ",coords)
print("xtras:       ",xtras)
print("file id:     ",fileid)
