""" 
========================================================================
|                                                                      |
|                              SUSY-AI                                 |
|   ----------------------------------------------------------------   |
|                           version 1.1.2                              |
|                           May 10th, 2016                             |
|                                                                      |
|                                                                      |
|   SUSY-AI is a powerful, yet simple interface between SUSY           |
|   data and the sklearn Python class. With the classifier that        |
|   comes with SUSY-AI, it allows for predicting of exclusion of a     |
|   datapoint in pMSSM parameterspace via sklearn using a              |
|   pre-trained classifier in only a couple of commands.               |
|                                                                      |
|   Full documentation, and the most recent version of the software    |
|   can be found on the project website                                |
|                                                                      |
|       https://susyai.hepforge.org/                                   |
|                                                                      |
|   If you are using this script scientifically or commercially,       |
|   please reference the following paper                               |
|                                                                      |
|       <tbd>                                                          |
|                                                                      |
|   For questions about the script, contact                            |
|                                                                      |
|       Bob Stienen <b.stienen@student.ru.nl>                          |
|                                                                      |
========================================================================
"""

__author__ = "Bob Stienen <b.stienen@student.ru.nl>"
__version__ = "1.1.2"


import sys
import os.path
import numpy as np
import time

try:
	import cPickle as pkl
except:
	import pickle as pkl

try:
	from sklearn.ensemble import RandomForestClassifier
except:
	raise ImportError("Python package sklearn has to be installed for SUSY-AI to work.")
	
try:
	import pyslha
except:
	print("")
	print("Warning: SUSY-AI needs the pyslha package to make predictions")
	print("         directly from spectrumfiles, but this package was not found.")
	print("         If this functionality is not used, SUSY-AI will work just fine.")
	print("         It is nevertheless advised to download this package. See")
	print("         documentation for details.")
	print("")


import pyslha


class susyai:
	bins = np.arange(0.005, 1.0, 0.01)
	confidence = [0.9987794955248169, 0.9947259208023517, 0.991825265040235, 0.990587817439176, 0.9841758241758242, 0.9818376068376068, 0.9730232558139534, 0.9727011494252874, 0.967400162999185, 0.9596622889305816, 0.9489124936772888, 0.9580218516388729, 0.9303452453058753, 0.9290969899665552, 0.9157458563535912, 0.9117221418234442, 0.9002320185614849, 0.9079263410728583, 0.8953068592057761, 0.9004484304932735, 0.8610567514677103, 0.8719165085388995, 0.8341968911917098, 0.8309716599190283, 0.8286038592508513, 0.8008519701810437, 0.8243093922651934, 0.7946537059538274, 0.8096885813148789, 0.7938496583143508, 0.7615112160566706, 0.77625, 0.7227214377406932, 0.7429718875502008, 0.7654155495978553, 0.7217847769028871, 0.7238805970149254, 0.7281045751633988, 0.695104895104895, 0.659846547314578, 0.6913099870298314, 0.639300134589502, 0.637741046831956, 0.5891364902506964, 0.5886426592797784, 0.6020408163265306, 0.6065789473684211, 0.5920096852300243, 0.5558441558441558, 0.5459317585301837, 0.568039950062422, 0.5026595744680851, 0.5129533678756477, 0.5020632737276478, 0.5099502487562189, 0.5573170731707318, 0.5300751879699248, 0.5643564356435644, 0.5530660377358491, 0.609375, 0.6169491525423729, 0.6182432432432432, 0.6071428571428572, 0.6495633187772926, 0.6568946796959827, 0.6728260869565217, 0.6809571286141576, 0.6870389884088515, 0.707876370887338, 0.7390917186108638, 0.7513761467889908, 0.761491760624458, 0.7803660565723793, 0.8170731707317074, 0.7905982905982906, 0.7983569828230022, 0.8183139534883721, 0.8492957746478873, 0.8420026007802341, 0.8424087123638693, 0.8650057937427578, 0.8806978052898142, 0.9006211180124224, 0.8980606663351567, 0.9192431933548685, 0.9218884120171674, 0.9251170046801872, 0.9238578680203046, 0.9385342789598109, 0.9477329974811083, 0.9542889390519187, 0.9594838709677419, 0.9687359424201529, 0.9702416616736569, 0.9768550034036759, 0.9848178137651822, 0.985517009093971, 0.991859577715594, 0.9958530459669411, 0.9988916348386131]
	mins = [-3999.99, -4000.0, 200.0, 90.879999, 91.419997, 90.02, 90.250001, 200.02999, 102.65, 200.00984, 264.01999, 200.0199, 100.06994, -7998.919, -3999.97, -3999.97, -4381.715, 1364.5547, 2.6724445]
	maxs = [3999.95, 3999.83, 3999.98, 4000.0, 3999.98, 3999.99, 4000.0, 3999.9899, 3999.9999, 3999.98, 3999.9999, 3999.97, 4000.0, 7993.83, 3999.95, 3999.94, 4132.4793, 37569812.0, 66.371989]
	varnames = ['M1', 'M2', 'M3', 'mL1', 'mL3', 'me1', 'me3', 'mQ1', 'mQt', 'mu1', 'mtR', 'md1', 'mbR', 'At', 'Ab', 'Atau', 'mu', 'mA2', 'tanbeta']

	def __init__(self, pickle_location = None, verbosity = 2, coordinate_selector = 1, xtra_selector = None, id_selector = None, map_outsiders = False):
		self.set_classifier()
		self.set_pickle_location(pickle_location)
		self.set_verbosity(verbosity)
		self.set_coordinate_selector(coordinate_selector)
		self.set_xtra_selector(xtra_selector)
		self.set_id_selector(id_selector)
		self.set_map_outsiders(map_outsiders)
		self.set_id_lists()
		self.set_given_warnings()
	
	""" ========================= Getters and setters ========================= """
	# Pickle location
	def get_pickle_location(self):
		return self.pickle_location
	def set_pickle_location(self, pickle_location = None):
		if pickle_location != None and not isinstance(pickle_location, str):
			raise TypeError("Supplied pickle location has to be a string")
		if pickle_location != None and not os.path.isfile(pickle_location):
			raise IOError("Pickle could not be found at indicated location")
		self.pickle_location = pickle_location
	# Verbosity
	def get_verbosity(self):
		return self.verbosity
	def set_verbosity(self, verbosity):
		if isinstance(verbosity, int) and verbosity in [0,1,2]:
			self.verbosity = int(verbosity)
		else:
			raise ValueError("Argument 'verbosity' in set_verbosity() must be in range(3). See documentation for details.")
	# Given warnings
	def get_given_warnings(self, name = None):
		if name == None:
			return self.given_warnings
		else:
			return name in self.get_given_warnings()
	def set_given_warnings(self, name = None):
		if name == None:
			self.given_warnings = []
		else:
			self.given_warnings.append(name)
	# ID lists
	def get_id_lists(self, name = None):
		if name == None:
			return self.id_lists
		else:
			if name in self.get_id_lists():
				return self.id_lists[name]
			else:
				return None
	def set_id_lists(self, name = None, fileid = None):
		if name == None:
			self.id_lists = {}
		else:
			if name not in self.get_id_lists():
				self.id_lists[name] = []
			if fileid != None:
				self.id_lists[name].append(fileid)
	# Coordinate selector
	def get_coordinate_selector(self):
		return self.csel
	def set_coordinate_selector(self, csel = 1):
		if isinstance(csel, int) or isinstance(csel, str):
			if csel == 1 or csel != "1":
				self.csel = [['MSOFT',1],['MSOFT',2],['MSOFT',3],['MSOFT',31],['MSOFT',33],['MSOFT',34],['MSOFT',36],['MSOFT',41],['MSOFT',43],['MSOFT',44],['MSOFT',46],['MSOFT',47],['MSOFT',49],['AU',(3,3)],['AD',(3,3)],['AE',(3,3)],['HMIX',1],['HMIX',4],['HMIX',2]]
			else:
				raise ValueError("Supplied version number for coordinate selection is not valid ('1' currently supported)")
		elif isinstance(csel, int) or isinstance(csel, str):
			self.csel = int(csel)
		elif isinstance(csel, list):
			if len(csel) == 19 and len(csel[0]) == 2:
				self.csel = csel
			else:
				raise ValueError("Supplied coordinate-selector has an unsupported shape, (19,2)-list expected.")
		elif hasattr(csel, '__call__'):
			self.csel = csel
		else:
			raise ValueError("Supplied argument for coordinate-selector is not valid. Supply a supported version number or a method (see documentation for details).")
	# Xtra selector
	def get_xtra_selector(self):
		return self.xsel
	def set_xtra_selector(self, xsel = None):
		if xsel == None or hasattr(xsel, '__call__'):
			self.xsel = xsel
		else:
			raise ValueError("Supplied argument for xtra-selector is not valid. Set it to None or supply a method (see documentation for details).")
	# ID selector
	def get_id_selector(self):
		return self.isel
	def set_id_selector(self, isel = None):
		if isel == None or hasattr(isel, '__call__'):
			self.isel = isel
		elif isinstance(isel, str) and (isel == 'filename' or isel == 'fullfilename'):
			self.isel = isel
		else:
			raise ValueError("Supplied argument for id-selector is not valid. Set it to None, 'filename' or 'fullfilename', or supply a method (see documentation for details).")
	# Map outsiders
	def get_map_outsiders(self):
		return self.map_outsiders
	def set_map_outsiders(self, map_outsiders = False):
			self.map_outsiders = bool(map_outsiders)
	# Classifier
	def get_classifier(self):
		return self.classifier
	def set_classifier(self, classifier = None):
		self.classifier = classifier
		
	""" ========================= Array interfaces ========================= """
	def get_min(self, i):
		if i in range(19):
			return self.mins[i]
		else:
			raise IndexError("get_min("+str(i)+") is out of range: argument must be in range(19)")	
	def get_max(self, i):
		if i in range(19):
			return self.maxs[i]
		else:
			raise IndexError("get_max("+str(i)+") is out of range: argument must be in range(19)")
	def get_name(self, i):
		if i in range(19):
			return self.varnames[i]
		else:
			raise IndexError("get_name("+str(i)+") is out of range: argument must be in range(19)")	

	""" ========================= Messages ========================= """
	def warn(self, text):
		if self.get_verbosity() > 0:
			self.message("\033[93m    WARNING   \033[0m"+str(text))
	def message(self, text):
		if self.get_verbosity() > 1:
			print(str(int(round(time.time())))+'   '+str(text))
	def warn_once(self, name, text):
		if self.get_verbosity() > 0:
			if not self.get_given_warnings(name):
				self.message("\033[93m    WARNING   \033[0m"+str(text))
				self.set_given_warnings(name)
			
	""" ========================= Pickle loader ========================= """
	def load_pickle(self):
		if self.pickle_location == None:
			raise Exception("No pickle was set to be loaded")
		self.set_classifier(RandomForestClassifier())
		with open(self.pickle_location, 'rb') as f:
			self.message("Loading pickle...")
			self.set_classifier(pkl.load(f))
	
	""" ========================= Select data ========================= """
	def get_spectrum(self, filepath):
		if not os.path.isfile(filepath):
			raise IOError("File '"+filepath+"' could not be found")
		return pyslha.read(filepath)
	@staticmethod
	def dict_to_matrix(dictionary):
		N = int(np.sqrt(len(dictionary)))
		matrix = np.zeros((N, N))
		for i in range(N):
			for j in range(N):
				matrix[i,j] = dictionary[i+1,j+1]
		return matrix
	def select_coordinates(self, spectrum):
		if hasattr(self.get_coordinate_selector(), '__call__'):
			return self.get_coordinate_selector()(spectrum)
		elif isinstance(self.get_coordinate_selector(), list):
			if len(self.get_coordinate_selector()) == 19 and len(self.get_coordinate_selector()[0]) == 2:
				coords = np.zeros(19)
				for i in range(19):
					coords[i] = spectrum.blocks[self.get_coordinate_selector()[i][0]][self.get_coordinate_selector()[i][1]]
				return coords
			else:
				raise ValueError("Supplied coordinate-selector has an unsupported shape, (19,2)-list expected.")
		else:
			raise ValueError("Supplied coordinate-selector was invalid. See documentation for details.")
	def select_xtras(self, spectrum, filepath):
		if self.get_xtra_selector() == None:
			return np.array([])
		elif hasattr(self.get_xtra_selector(), '__call__'):
			return self.get_xtra_selector()(spectrum, filepath)
		else:
			raise ValueError("Supplied xtra-selector was invalid. See documentation for details.")
	def select_id(self, spectrum, filepath):
		if self.get_id_selector() == None:
			return None
		elif hasattr(self.get_id_selector(), '__call__'):
			return self.get_id_selector()(spectrum, filepath)
		elif self.get_id_selector() == 'filename':
			f = filepath.split('/')[-1]
			if '.' in f:
				f = '.'.join(f.split('.')[:-1])
			return f
		elif self.get_id_selector() == 'fullfilename':
			return filepath.split('/')[-1]
		else:
			raise ValueError("Supplied xtra-selector was invalid. See documentation for details.")
	def check_lsp(self, fileid, spectrum):
		if 'MASS' in spectrum.blocks:
			masses = np.array([[i, spectrum.blocks['MASS'][i]] for i in spectrum.blocks['MASS'].keys() if i > 1e6])
			masses = masses[abs(masses[:,1]).argsort()]
			if masses[0,0] != 1000022 and 1000022 in masses[:,0]:
				if self.get_id_selector() != None:
					self.warn_once("LSP01", "File detected where ~chi01 is not the LSP. Prediction will continue, but will possibly be incorrect.")
					self.warn_once("LSP02", "All ids of the files for which this holds will be saved internally.")
					self.warn_once("LSP03", "These ids can be retrieved by calling get_id_lists('LSP').")
					self.set_id_lists('LSP',fileid)
				else:
					self.warn_once("LSP04", "File detected where ~chi01 is not the LSP. Prediction will continue, but will possibly be incorrect.")
					self.warn_once("LSP05", "No id selector was specified. If it were, all ids of the files for which this holds would be saved internally.")
					self.warn_once("LSP06", "These ids can be retrieved by calling get_id_lists('LSP').")
		else:
			self.warn_once("LSP07","LSP could not be identified due to absence of the MASS block or the mass of the lightest neutralino in the spectrumfile.")
			self.warn_once("LSP08","SUSY-AI assumes ~chi01 to be the LSP. Prediction will continue, but is possibly incorrect.")
	def get_data_from_file(self, filepath):
		spectrum = self.get_spectrum(filepath)
		coordinates = self.select_coordinates(spectrum)
		xtras = self.select_xtras(spectrum, filepath)
		fileid = self.select_id(spectrum, filepath)
		self.check_lsp(fileid, spectrum)
		return (coordinates, xtras, fileid)
	
	""" ========================= Predict ========================= """
	def map_coordinates(self, coordinates):
		for i in range(19):
			r = self.maxs[i] - self.mins[i]
			newmax = (self.maxs[i] - 0.1125*r)
			coordinates[coordinates[:,i] > newmax, i] = newmax
			if self.mins[i] < 0:
				newmin = (self.mins[i] + 0.1125*r)
				coordinates[coordinates[:,i] < newmin, i] = newmin
		return coordinates
	def calculate_results(self, prediction):
		confidence = np.zeros(len(prediction))
		for i in range(len(prediction)):
			b = np.argmin(abs(self.bins - prediction[i]))
			confidence[i] = self.confidence[b]
		classification = (prediction > self.bins[np.argmin(self.confidence)])*1.0
		return (classification, confidence)
	def apply_min_confidence(self, arrayorlist, confidence, min_confidence):
		if len(arrayorlist) != len(confidence):
			raise ValueError("Lengths of arrayorlist and confidence have to be equal.")
		if type(arrayorlist) != np.ndarray and type(arrayorlist) != list:
			raise ValueError("Inputparameter arrayorlist has to be an array or a list.")
		r = [arrayorlist[i] for i in range(len(arrayorlist)) if confidence[i] >= min_confidence]
		if type(arrayorlist) == np.ndarray:
			r = np.array(r)
		return r
	def predict(self, coordinates, min_confidence = 0.5):
		if min_confidence < 0 and min_confidence > 1.0:
			raise ValueError("Only required certainties between 0 and 1.0 are allowed.")
		if self.get_classifier() == None:
			self.warn("If load_pickle() is not used before using predict(), loading is part of prediction and will therefore be considerably slowed down.")
			self.load_pickle()
		if self.get_map_outsiders() == True:
			self.message("Mapping extreme coordinates...")
			coordinates = self.map_coordinates(coordinates)
		self.message("Predicting...")
		if len(coordinates.shape) == 1:
			coordinates = coordinates.reshape(1,-1)
		prediction = self.get_classifier().predict_proba( coordinates )[:,1]
		self.message("Expanding results...")
		classification, confidence = self.calculate_results(prediction)
		if min_confidence > 0.5:
			self.message("Applying required confidence level...")
			L_orig = len(classification)
			classification = self.apply_min_confidence(classification, confidence, min_confidence)
			prediction = self.apply_min_confidence(prediction, confidence, min_confidence)
			coordinates = self.apply_min_confidence(coordinates, confidence, min_confidence)
			confidence = self.apply_min_confidence(confidence, confidence, min_confidence)
			self.message("   datapoints before: "+str(L_orig))
			self.message("   datapoints after:  "+str(len(confidence)))
		self.message("Prediction finished, result returned...")
		return (classification, prediction, confidence, coordinates)
	
	""" ========================= Predict files ========================= """
	def predict_files(self, filepaths, min_confidence = 0.5):
		N = len(filepaths)
		coordinates = np.zeros((N, 19))
		xtras = None
		fileids = [None] * N
		self.message("Clear memory of previous runs (if any)...")
		self.set_given_warnings()
		self.set_id_lists()
		self.message("Loading files...")
		for i in range(N):
			c, x, f = self.get_data_from_file(filepaths[i])
			coordinates[i,:] = c
			if self.get_xtra_selector() != None:
				if xtras == None:
					xtras = np.zeros((N,len(x)))
				xtras[i,:] = x
			if self.get_id_selector() != None:
				fileids[i] = f
		classification, prediction, confidence, coordinates = self.predict(coordinates)
		
		if min_confidence > 0.5:
			self.message("Applying required confidence level...")
			L_orig = len(classification)
			classification = self.apply_min_confidence(classification, confidence, min_confidence)
			prediction = self.apply_min_confidence(prediction, confidence, min_confidence)
			coordinates = self.apply_min_confidence(coordinates, confidence, min_confidence)
			if self.get_xtra_selector() != None:
				xtras = self.apply_min_confidence(xtras, confidence, min_confidence)
			if self.get_id_selector() != None:
				fileids = self.apply_min_confidence(fileids, confidence, min_confidence)
			confidence = self.apply_min_confidence(confidence, confidence, min_confidence)
			self.message("   datapoints before: "+str(L_orig))
			self.message("   datapoints after:  "+str(len(confidence)))
		
		self.message("Map results...")
		result = (classification, prediction, confidence, coordinates)
		if self.get_xtra_selector() != None:
			result = result + (xtras,)
		if self.get_id_selector() != None:
			result = result + (np.array(fileids),)
		return result
